/* Copyright (C) 2005  David Decotigny

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA. 
*/
#include <sos/thread.h>
#include <sos/kmalloc.h>
#include <sos/klibc.h>
#include <drivers/bochs.h>
#include <hwcore/paging.h>
#include <sos/physmem.h>
#include <sos/umem_vmm.h>
#include <drivers/zero.h>
#include <drivers/mem.h>
#include <sos/binfmt_elf32.h>

#include <hwcore/cpu_context.h>
#include <sos/uaccess.h>
#include "syscall.h"


/**
 * THE syscall entry point
 */
sos_ret_t sos_do_syscall(int syscall_id,
			 const struct sos_cpu_state *user_ctxt)
{
  sos_ret_t retval;

  switch(syscall_id)
    {
    case SOS_SYSCALL_ID_EXIT:
      {
	unsigned int status;

	retval = sos_syscall_get1arg(user_ctxt, & status);
	if (SOS_OK != retval)
	  break;
	sos_thread_exit();
	retval = -SOS_EFATAL; /* Not reached */
      }
      break;

    case SOS_SYSCALL_ID_FORK:
      {
	struct sos_thread *cur_thr, *new_thr;
	struct sos_process *new_proc;

	cur_thr = sos_thread_get_current();

	/* Duplicate the current process (and its address space) */
	new_proc = sos_process_create(NULL, TRUE);
	if (! new_proc)
	  {
	    retval = -SOS_ENOMEM;
	    break;
	  }
	
	/* Create *the* thread in this new processs, copy of the
	   current user thread (same registers, EXCEPT eax which is
	   set to 0) */
	new_thr =
	  sos_duplicate_user_thread(NULL, new_proc,
				    cur_thr,
				    user_ctxt,
				    0);
	if (! new_thr)
	  {
	    sos_process_unref(new_proc);
	    retval = -SOS_ENOMEM;
	    break;
	  }

	sos_process_unref(new_proc);

	/* Return to the "parent" thread with a value different from
	   0. Unix says it should be the "PID" of the child. We don't
	   have such a "PID" notion for now */
	retval = (sos_ui32_t)new_proc;
      }
      break;

    case SOS_SYSCALL_ID_EXEC:
      {
	struct sos_thread *cur_thr, *new_thr;
	struct sos_process * proc;
	struct sos_umem_vmm_as *new_as;
	sos_uaddr_t user_str, ustack, start_uaddr;
	sos_size_t len;
	char * str;

	cur_thr = sos_thread_get_current();
	proc    = cur_thr->process;

	/* Make sure the process has exactly 1 thread in it */
	if (sos_process_get_nb_threads(proc) != 1)
	  {
	    retval = -SOS_EBUSY;
	    break;
	  }

	/* Get the user arguments */
	retval = sos_syscall_get2args(user_ctxt, & user_str, & len);
	if (SOS_OK != retval)
	  break;

	/* Copy the program name into kernel sppace */
	retval = sos_strndup_from_user(& str, user_str, len + 1, 0);
	if (SOS_OK != retval)
	  {
	    break;
	  }

	/* Create a new empty address space to map the program */
	new_as = sos_umem_vmm_create_empty_as(proc);
	if (! new_as)
	  {
	    sos_kfree((sos_vaddr_t)str);
	    retval = -SOS_ENOMEM;
	    break;
	  }

	/* Map the program in it */
	start_uaddr = sos_binfmt_elf32_map(new_as, str);
	if (start_uaddr == (sos_uaddr_t)NULL)
	  {
	    sos_umem_vmm_delete_as(new_as);
	    sos_kfree((sos_vaddr_t)str);
	    retval = -SOS_ENOENT;
	    break;
	  }

	/* Allocate space for the user stack (8MB) */
#define SOS_DEFAULT_USER_STACK_SIZE (8 << 20)
	ustack = (SOS_PAGING_TOP_USER_ADDRESS - SOS_DEFAULT_USER_STACK_SIZE)
	            + 1;
	retval = sos_dev_zero_map(new_as, &ustack, SOS_DEFAULT_USER_STACK_SIZE,
				  SOS_VM_MAP_PROT_READ | SOS_VM_MAP_PROT_WRITE,
				  /* PRIVATE */ 0);
	if (SOS_OK != retval)
	  {
	    sos_umem_vmm_delete_as(new_as);
	    sos_kfree((sos_vaddr_t)str);
	    break;
	  }

	/* Now create the user thread */
	new_thr = sos_create_user_thread(NULL,
					 proc,
					 start_uaddr,
					 0, 0,
					 ustack + SOS_DEFAULT_USER_STACK_SIZE
					   - 4,
					 SOS_SCHED_PRIO_TS_LOWEST);
	if (! new_thr)
	  {
	    sos_umem_vmm_delete_as(new_as);
	    sos_kfree((sos_vaddr_t)str);
	    retval = -SOS_ENOMEM;
	    break;	    
	  }

	sos_process_set_name(proc, str);

	/* Switch to this address space */
	retval = sos_process_set_address_space(proc,
					       new_as);
	if (SOS_OK != retval)
	  {
	    sos_umem_vmm_delete_as(new_as);
	    sos_kfree((sos_vaddr_t)str);
	    break;	    
	  }

	/* The current thread must exit now */
	sos_kfree((sos_vaddr_t)str);
	sos_thread_exit();
	retval = -SOS_EFATAL;
      }
      break;

    case SOS_SYSCALL_ID_FAKEMMAP:
      {
	sos_uaddr_t ptr_hint_uaddr;
	sos_uaddr_t hint_uaddr;
	sos_size_t  length;
	sos_ui32_t  prot;
	sos_ui32_t  flags;
	sos_uaddr_t name_user;
	sos_ui32_t  offs64_hi;
	sos_ui32_t  offs64_lo;
	sos_luoffset_t offset_in_resource;
	char        name[256];
	struct sos_umem_vmm_as * my_as;
	
	retval = sos_syscall_get7args(user_ctxt,
				      (unsigned int*)& ptr_hint_uaddr,
				      (unsigned int*)& length,
				      (unsigned int*)& prot,
				      (unsigned int*)& flags,
				      (unsigned int*)& name_user,
				      (unsigned int*)& offs64_hi,
				      (unsigned int*)& offs64_lo);
	if (SOS_OK != retval)
	  break;

	/* Compute 64 bits offset value */
	offset_in_resource   = offs64_hi;
	offset_in_resource <<= 32;
	offset_in_resource  |= offs64_lo;

	retval = sos_memcpy_from_user((sos_vaddr_t)& hint_uaddr,
				      ptr_hint_uaddr,
				      sizeof(hint_uaddr));
	if (sizeof(hint_uaddr) != retval)
	  {
	    retval = -SOS_EFAULT;
	    break;
	  }

	retval = sos_strzcpy_from_user(name, name_user, sizeof(name));
	if (SOS_OK != retval)
	  break;

	my_as
	  = sos_process_get_address_space(sos_thread_get_current()->process);
	if ( (0 == strncmp(name, "/dev/zero", sizeof(name)))
	     || (0 == strncmp(name, "/dev/null", sizeof(name))) )
	  retval = sos_dev_zero_map(my_as, & hint_uaddr, length, prot, flags);
	else if (0 == strncmp(name, "/dev/mem", sizeof(name)))
	  retval = sos_dev_physmem_map(my_as, & hint_uaddr, length,
				       offset_in_resource, prot, flags);
	else if (0 == strncmp(name, "/dev/kmem", sizeof(name)))
	  retval = sos_dev_kmem_map(my_as, & hint_uaddr, length,
				    offset_in_resource, prot, flags);
	else
	  retval = -SOS_ENOENT;

	if (SOS_OK == retval)
	  {
	    if (sizeof(hint_uaddr)
		!= sos_memcpy_to_user(ptr_hint_uaddr,
				      (sos_vaddr_t)& hint_uaddr,
				      sizeof(hint_uaddr)))
	      {
		sos_umem_vmm_unmap(my_as, hint_uaddr, length);
		retval = -SOS_EFAULT;
	      }
	  }

      }
      break;

    case SOS_SYSCALL_ID_MUNMAP:
      {
	sos_uaddr_t start_uaddr;
	sos_size_t  size;
	struct sos_umem_vmm_as * my_as;

	my_as
	  = sos_process_get_address_space(sos_thread_get_current()->process);

	retval = sos_syscall_get2args(user_ctxt,
				      (unsigned int*)& start_uaddr,
				      (unsigned int*)& size);
	if (SOS_OK != retval)
	  break;

	retval = sos_umem_vmm_unmap(my_as, start_uaddr, size);
      }
      break;

    case SOS_SYSCALL_ID_MPROTECT:
      {
	sos_uaddr_t start_uaddr;
	sos_size_t  size;
	sos_ui32_t  new_access_rights;
	struct sos_umem_vmm_as * my_as;

	my_as
	  = sos_process_get_address_space(sos_thread_get_current()->process);

	retval = sos_syscall_get3args(user_ctxt,
				      (unsigned int*)& start_uaddr,
				      (unsigned int*)& size,
				      (unsigned int*)& new_access_rights);
	if (SOS_OK != retval)
	  break;

	retval = sos_thread_prepare_user_space_access(NULL, (sos_vaddr_t)NULL);
	if (SOS_OK != retval)
	  break;

	retval = sos_umem_vmm_chprot(my_as, start_uaddr, size,
				     new_access_rights);

	sos_thread_end_user_space_access();
      }
      break;

    case SOS_SYSCALL_ID_MRESIZE:
      {
	sos_uaddr_t old_uaddr;
	sos_size_t  old_size;
	sos_uaddr_t *uptr_new_uaddr;
	sos_uaddr_t new_uaddr;
	sos_size_t  new_size;
	sos_ui32_t  flags;
	struct sos_umem_vmm_as * my_as;

	my_as
	  = sos_process_get_address_space(sos_thread_get_current()->process);

	retval = sos_syscall_get5args(user_ctxt,
				      (unsigned int*)& old_uaddr,
				      (unsigned int*)& old_size,
				      (unsigned int*)& uptr_new_uaddr,
				      (unsigned int*)& new_size,
				      (unsigned int*)& flags);
	if (SOS_OK != retval)
	  break;

	if (sizeof(new_uaddr) != sos_memcpy_from_user((sos_vaddr_t)& new_uaddr,
						      (sos_uaddr_t)
						        uptr_new_uaddr,
						      sizeof(new_uaddr)))
	  {
	    retval = -SOS_EFAULT;
	    break;
	  }

	retval = sos_thread_prepare_user_space_access(NULL, (sos_vaddr_t)NULL);
	if (SOS_OK != retval)
	  break;

	retval = sos_umem_vmm_resize(my_as, old_uaddr, old_size,
				     & new_uaddr, new_size, flags);
	sos_thread_end_user_space_access();
	if (SOS_OK != retval)
	  break;

	if (sizeof(new_uaddr)
	    == sos_memcpy_to_user((sos_uaddr_t)uptr_new_uaddr,
				  (sos_vaddr_t)&new_uaddr,
				  sizeof(new_uaddr)))
	  {
	    retval = -SOS_EFAULT;
	    break;
	  }
      }
      break;

    case SOS_SYSCALL_ID_NEW_THREAD:
      {
	sos_uaddr_t start_func;
	sos_ui32_t  start_arg1, start_arg2;
	sos_size_t  stack_size;
	sos_uaddr_t stack_uaddr;

	struct sos_thread * new_thr;
	struct sos_umem_vmm_as * my_as;

	my_as
	  = sos_process_get_address_space(sos_thread_get_current()->process);

	retval = sos_syscall_get4args(user_ctxt,
				      (unsigned int*)& start_func,
				      (unsigned int*)& start_arg1,
				      (unsigned int*)& start_arg2,
				      (unsigned int*)& stack_size);
	if (SOS_OK != retval)
	  break;

	if (stack_size <= 0)
	  {
	    retval = -SOS_EINVAL;
	    break;
	  }

	/* Allocate the stack */
	stack_uaddr = 0;
	stack_size = SOS_PAGE_ALIGN_SUP(stack_size);
	retval = sos_dev_zero_map(my_as, & stack_uaddr, stack_size,
				  SOS_VM_MAP_PROT_READ | SOS_VM_MAP_PROT_WRITE,
				  /* PRIVATE */ 0);
	if (SOS_OK != retval)
	  break;

	/* Now create the user thread */
	new_thr = sos_create_user_thread(NULL,
					 sos_thread_get_current()->process,
					 start_func,
					 start_arg1,
					 start_arg2,
					 stack_uaddr + stack_size - 4,
					 SOS_SCHED_PRIO_TS_LOWEST);

	if (! new_thr)
	  {
	    sos_umem_vmm_unmap(my_as, stack_uaddr, stack_size);
	    retval = -SOS_ENOMEM;
	    break;	    
	  }

      }
      break;

    case SOS_SYSCALL_ID_NANOSLEEP:
      {
	struct sos_time delay;

	retval = sos_syscall_get2args(user_ctxt,
				      (unsigned int*)& delay.sec,
				      (unsigned int*)& delay.nanosec);
	if (SOS_OK != retval)
	  break;

	retval = sos_thread_sleep(& delay);
      }
      break;

    case SOS_SYSCALL_ID_BRK:
      {
	sos_uaddr_t new_top_heap;
	struct sos_umem_vmm_as * my_as;

	my_as
	  = sos_process_get_address_space(sos_thread_get_current()->process);

	retval = sos_syscall_get1arg(user_ctxt,
				     (unsigned int*)& new_top_heap);
	if (SOS_OK != retval)
	  break;

	retval = sos_thread_prepare_user_space_access(NULL, (sos_vaddr_t)NULL);
	if (SOS_OK != retval)
	  break;

	retval = sos_umem_vmm_brk(my_as, new_top_heap);
	sos_thread_end_user_space_access();
      }
      break;

      
      /**
       * File system interface
       */
    case SOS_SYSCALL_ID_MOUNT:
      {
	sos_uaddr_t user_src;
	sos_size_t srclen;
	const char * kernel_src = NULL;
	sos_uaddr_t user_dst;
	sos_size_t dstlen;
	const char * kernel_dst;
	sos_ui32_t mountflags;
	sos_uaddr_t user_fstype;
	char * kernel_fstype;
	sos_uaddr_t user_args;
	char * kernel_args = NULL;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get7args(user_ctxt,
				      (unsigned int*)& user_src,
				      (unsigned int*)& srclen,
				      (unsigned int*)& user_dst,
				      (unsigned int*)& dstlen,
				      (unsigned int*)& user_fstype,
				      (unsigned int*)& mountflags,
				      (unsigned int*)& user_args);
	if (SOS_OK != retval)
	  break;

	if (user_src != (sos_uaddr_t)NULL)
	  {
	    retval = sos_memdup_from_user((sos_vaddr_t*) &kernel_src, user_src, srclen, 0);
	    if (SOS_OK != retval)
	      break;
	  }

	retval = sos_memdup_from_user((sos_vaddr_t*) &kernel_dst, user_dst, dstlen, 0);
	if (SOS_OK != retval)
	  {
	    if (kernel_src)
	      sos_kfree((sos_vaddr_t)kernel_src);
	    break;
	  }

	retval = sos_strndup_from_user(& kernel_fstype, user_fstype, 256, 0);
	if (SOS_OK != retval)
	  {
	    if (kernel_src)
	      sos_kfree((sos_vaddr_t)kernel_src);
	    sos_kfree((sos_vaddr_t)kernel_dst);
	    break;
	  }

	if (user_args != (sos_uaddr_t)NULL)
	  {
	    retval = sos_strndup_from_user(& kernel_args, user_args, 1024, 0);
	    if (SOS_OK != retval)
	      {
		if (kernel_src)
		  sos_kfree((sos_vaddr_t)kernel_src);
		sos_kfree((sos_vaddr_t)kernel_dst);
		sos_kfree((sos_vaddr_t)kernel_fstype);
		break;
	      }
	  }

	retval = sos_fs_mount(proc, kernel_src, srclen,
			      kernel_dst, dstlen,
			      kernel_fstype,
			      mountflags,
			      kernel_args,
			      NULL);
	if (kernel_src)
	  sos_kfree((sos_vaddr_t)kernel_src);
	sos_kfree((sos_vaddr_t)kernel_dst);
	sos_kfree((sos_vaddr_t)kernel_fstype);
	if (kernel_args)
	  sos_kfree((sos_vaddr_t)kernel_args);
      }
      break;

    case SOS_SYSCALL_ID_UMOUNT:
      {
	sos_uaddr_t user_str;
	sos_size_t  len;
	char * path;
	struct sos_process * proc;
	
	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get2args(user_ctxt,
				      (unsigned int*)& user_str,
				      (unsigned int*)& len);
	if (SOS_OK != retval)
	  break;
	
	retval = sos_memdup_from_user((sos_vaddr_t*) &path, user_str, len, 0);
	if (SOS_OK != retval)
	  break;
	
	retval = sos_fs_umount(proc,
			       path, len);
	sos_kfree((sos_vaddr_t)path);
      }
      break;

    case SOS_SYSCALL_ID_SYNC:
      {
	sos_fs_sync_all_fs();
	retval = SOS_OK;
      }
      break;

    case SOS_SYSCALL_ID_VFSTAT64:
      {
	sos_uaddr_t user_str;
	sos_size_t  len;
	sos_uaddr_t user_vfstat_struct;
	struct sos_fs_statfs kernel_vfstat_struct;
	char * path;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get3args(user_ctxt,
				      (unsigned int*)& user_str,
				      (unsigned int*)& len,
				      (unsigned int*)& user_vfstat_struct);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &path, user_str, len, 0);
	if (SOS_OK != retval)
	  break;

	retval = sos_fs_vfstat(proc, path, len, & kernel_vfstat_struct);
	sos_kfree((sos_vaddr_t)path);
	if (SOS_OK != retval)
	  break;

	if (sizeof(kernel_vfstat_struct)
	    != sos_memcpy_to_user(user_vfstat_struct,
				  (sos_vaddr_t) & kernel_vfstat_struct,
				  sizeof(kernel_vfstat_struct)))
	  retval = -SOS_EFAULT;
      }
      break;

    case SOS_SYSCALL_ID_OPEN:
      {
	sos_uaddr_t user_str;
	sos_size_t  len;
	sos_ui32_t  open_flags;
	sos_ui32_t  access_rights;
	char * path;
	struct sos_fs_opened_file * of;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get4args(user_ctxt,
				      (unsigned int*)& user_str,
				      (unsigned int*)& len,
				      (unsigned int*)& open_flags,
				      (unsigned int*)& access_rights);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &path, user_str, len, 0);
	if (SOS_OK != retval)
	  break;

	retval = sos_fs_open(proc,
			     path, len,
			     open_flags,
			     access_rights,
			     & of);
	sos_kfree((sos_vaddr_t)path);
	if (SOS_OK != retval)
	  break;

	retval = sos_process_register_opened_file(proc, of);
	if (retval < 0)
	  {
	    sos_fs_close(of);
	    break;
	  }
      }
      break;

    case SOS_SYSCALL_ID_CLOSE:
      {
	struct sos_fs_opened_file * of;
	struct sos_process * proc;
	int fd;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get1arg(user_ctxt,
				     (unsigned int*)& fd);
	if (SOS_OK != retval)
	  break;

	of = sos_process_get_opened_file(proc, fd);
	if (NULL == of)
	  {
	    retval = -SOS_EBADF;
	    break;
	  }

	retval = sos_process_unregister_opened_file(proc, fd);
	if (SOS_OK != retval)
	  break;

	retval = sos_fs_close(of);
      }
      break;

    case SOS_SYSCALL_ID_READ:
      {
	struct sos_fs_opened_file * of;
	struct sos_process * proc;
	sos_uaddr_t uaddr_buf;
	sos_uaddr_t uaddr_buflen;
	sos_size_t kernel_buflen;
	int fd;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get3args(user_ctxt,
				      (unsigned int*)& fd,
				      (unsigned int*)& uaddr_buf,
				      (unsigned int*)& uaddr_buflen);
	if (SOS_OK != retval)
	  break;

	/* Retrieve the value for "buflen" */
	retval = sos_memcpy_from_user((sos_vaddr_t)& kernel_buflen,
				      uaddr_buflen,
				      sizeof(kernel_buflen));
	if (sizeof(kernel_buflen) != retval)
	  {
	    retval = -SOS_EFAULT;
	    break;
	  }

	of = sos_process_get_opened_file(proc, fd);
	if (NULL == of)
	  {
	    retval = -SOS_EBADF;
	    break;
	  }

	/* Do the actual reading */
	retval = sos_fs_read(of, uaddr_buf, & kernel_buflen);
	
	/* Send successful number of bytes read to user */
	sos_memcpy_to_user(uaddr_buflen,
			   (sos_vaddr_t)& kernel_buflen,
			   sizeof(kernel_buflen));
      }
      break;

    case SOS_SYSCALL_ID_READDIR:
      {
	struct sos_fs_opened_file * of;
	struct sos_process * proc;
	sos_uaddr_t uaddr_direntry;
	struct sos_fs_dirent direntry;
	int fd;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get2args(user_ctxt,
				      (unsigned int*)& fd,
				      (unsigned int*)& uaddr_direntry);
	if (SOS_OK != retval)
	  break;

	of = sos_process_get_opened_file(proc, fd);
	if (NULL == of)
	  {
	    retval = -SOS_EBADF;
	    break;
	  }

	/* Do the actual readdir */
	retval = sos_fs_readdir(of, & direntry);
	if (SOS_OK != retval)
	  break;
	
	/* Send direntry structure to user */
	if (sizeof(direntry) != sos_memcpy_to_user(uaddr_direntry,
						   (sos_vaddr_t)& direntry,
						   sizeof(direntry)))
	  retval = -SOS_EFAULT;
      }
      break;

    case SOS_SYSCALL_ID_WRITE:
      {
	struct sos_fs_opened_file * of;
	struct sos_process * proc;
	sos_uaddr_t uaddr_buf;
	sos_uaddr_t uaddr_buflen;
	sos_size_t kernel_buflen;
	int fd;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get3args(user_ctxt,
				      (unsigned int*)& fd,
				      (unsigned int*)& uaddr_buf,
				      (unsigned int*)& uaddr_buflen);
	if (SOS_OK != retval)
	  break;

	/* Retrieve the value for "buflen" */
	retval = sos_memcpy_from_user((sos_vaddr_t)& kernel_buflen,
				      uaddr_buflen,
				      sizeof(kernel_buflen));
	if (sizeof(kernel_buflen) != retval)
	  {
	    retval = -SOS_EFAULT;
	    break;
	  }

	of = sos_process_get_opened_file(proc, fd);
	if (NULL == of)
	  {
	    retval = -SOS_EBADF;
	    break;
	  }

	/* Do the actual writing */
	retval = sos_fs_write(of, uaddr_buf, & kernel_buflen);
	
	/* Send successful number of bytes written to user */
	sos_memcpy_to_user(uaddr_buflen,
			   (sos_vaddr_t)& kernel_buflen,
			   sizeof(kernel_buflen));
      }
      break;

    case SOS_SYSCALL_ID_SEEK64:
      {
	struct sos_fs_opened_file * of;
	struct sos_process * proc;
	sos_uaddr_t uaddr_offset;
	sos_seek_whence_t whence;
	sos_lsoffset_t kernel_offset, result_position;
	int fd;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get3args(user_ctxt,
				      (unsigned int*)& fd,
				      (unsigned int*)& uaddr_offset,
				      (unsigned int*)& whence);
	if (SOS_OK != retval)
	  break;

	/* Retrieve the value for "buflen" */
	retval = sos_memcpy_from_user((sos_vaddr_t)& kernel_offset,
				      uaddr_offset,
				      sizeof(kernel_offset));
	if (sizeof(kernel_offset) != retval)
	  {
	    retval = -SOS_EFAULT;
	    break;
	  }

	of = sos_process_get_opened_file(proc, fd);
	if (NULL == of)
	  {
	    retval = -SOS_EBADF;
	    break;
	  }

	/* Do the actual seek */
	retval = sos_fs_seek(of, kernel_offset, whence, & result_position);
	
	/* Send successful number of bytes written to user */
	sos_memcpy_to_user(uaddr_offset,
			   (sos_vaddr_t)& result_position,
			   sizeof(kernel_offset));
      }
      break;

    case SOS_SYSCALL_ID_FTRUNCATE64:
      {
	struct sos_fs_opened_file * of;
	struct sos_process * proc;
	sos_lsoffset_t length;
	int fd;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get2args(user_ctxt,
				      (unsigned int*)& fd,
				      (unsigned int*)& length);
	if (SOS_OK != retval)
	  break;

	of = sos_process_get_opened_file(proc, fd);
	if (NULL == of)
	  {
	    retval = -SOS_EBADF;
	    break;
	  }

	/* Do the actual trunc */
	retval = sos_fs_ftruncate(of, length);
      }
      break;

    case SOS_SYSCALL_ID_FSMMAP:
      {
	sos_uaddr_t ptr_hint_uaddr;
	sos_uaddr_t hint_uaddr;
	sos_size_t  length;
	sos_ui32_t  prot;
	sos_ui32_t  flags;
	int         fd;
	sos_ui32_t  offs64_hi;
	sos_ui32_t  offs64_lo;
	sos_luoffset_t offset_in_resource;
	struct sos_fs_opened_file * of;
	struct sos_process * proc;
	
	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get7args(user_ctxt,
				      (unsigned int*)& ptr_hint_uaddr,
				      (unsigned int*)& length,
				      (unsigned int*)& prot,
				      (unsigned int*)& flags,
				      (unsigned int*)& fd,
				      (unsigned int*)& offs64_hi,
				      (unsigned int*)& offs64_lo);
	if (SOS_OK != retval)
	  break;

	of = sos_process_get_opened_file(proc, fd);
	if (NULL == of)
	  {
	    retval = -SOS_EBADF;
	    break;
	  }

	/* Compute 64 bits offset value */
	offset_in_resource   = offs64_hi;
	offset_in_resource <<= 32;
	offset_in_resource  |= offs64_lo;

	retval = sos_memcpy_from_user((sos_vaddr_t)& hint_uaddr,
				      ptr_hint_uaddr,
				      sizeof(hint_uaddr));
	if (sizeof(hint_uaddr) != retval)
	  {
	    retval = -SOS_EFAULT;
	    break;
	  }

	retval = sos_fs_mmap(of, & hint_uaddr, length, prot, flags,
			     offset_in_resource);
	if (SOS_OK == retval)
	  {
	    if (sizeof(hint_uaddr)
		!= sos_memcpy_to_user(ptr_hint_uaddr,
				      (sos_vaddr_t)& hint_uaddr,
				      sizeof(hint_uaddr)))
	      {
		sos_umem_vmm_unmap(sos_process_get_address_space(proc),
				   hint_uaddr, length);
		retval = -SOS_EFAULT;
	      }
	  }

      }
      break;

    case SOS_SYSCALL_ID_FSYNC:
      {
	struct sos_fs_opened_file * of;
	struct sos_process * proc;
	int fd;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get1arg(user_ctxt,
				     (unsigned int*)& fd);
	if (SOS_OK != retval)
	  break;

	of = sos_process_get_opened_file(proc, fd);
	if (NULL == of)
	  {
	    retval = -SOS_EBADF;
	    break;
	  }

	/* Do the actual sync */
	retval = sos_fs_fsync(of);
      }
      break;

    case SOS_SYSCALL_ID_FCNTL:
      {
	struct sos_fs_opened_file * of;
	struct sos_process * proc;
	sos_ui32_t cmd, arg;
	int fd;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get3args(user_ctxt,
				      (unsigned int*)& fd,
				      (unsigned int*)& cmd,
				      (unsigned int*)& arg);
	if (SOS_OK != retval)
	  break;

	of = sos_process_get_opened_file(proc, fd);
	if (NULL == of)
	  {
	    retval = -SOS_EBADF;
	    break;
	  }

	/* Do the actual fcntl */
	retval = sos_fs_fcntl(of, cmd, arg);
      }
      break;

    case SOS_SYSCALL_ID_CREAT:
      {
	sos_uaddr_t user_str;
	sos_size_t  len;
	sos_ui32_t  access_rights;
	char * path;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get3args(user_ctxt,
				      (unsigned int*)& user_str,
				      (unsigned int*)& len,
				      (unsigned int*)& access_rights);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &path, user_str, len, 0);
	if (SOS_OK != retval)
	  break;

	retval = sos_fs_creat(proc,
			      path, len,
			      access_rights);
	sos_kfree((sos_vaddr_t)path);
      }
      break;

    case SOS_SYSCALL_ID_LINK:
    case SOS_SYSCALL_ID_RENAME:
      {
	sos_uaddr_t user_oldpath, user_newpath;
	sos_size_t  oldpathlen, newpathlen;
	char * kernel_oldpath, * kernel_newpath;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get4args(user_ctxt,
				      (unsigned int*)& user_oldpath,
				      (unsigned int*)& oldpathlen,
				      (unsigned int*)& user_newpath,
				      (unsigned int*)& newpathlen);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &kernel_oldpath,
				      user_oldpath,
				      oldpathlen, 0);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &kernel_newpath,
				      user_newpath,
				      newpathlen, 0);
	if (SOS_OK != retval)
	  {
	    sos_kfree((sos_vaddr_t) kernel_oldpath);
	    break;
	  }

	if (syscall_id == SOS_SYSCALL_ID_LINK)
	  retval = sos_fs_link(proc,
			       kernel_oldpath, oldpathlen,
			       kernel_newpath, newpathlen);
	else
	  retval = sos_fs_rename(proc,
				 kernel_oldpath, oldpathlen,
				 kernel_newpath, newpathlen);
	sos_kfree((sos_vaddr_t)kernel_oldpath);
	sos_kfree((sos_vaddr_t)kernel_newpath);
      }
      break;

    case SOS_SYSCALL_ID_UNLINK:
      {
	sos_uaddr_t user_str;
	sos_size_t  len;
	char * path;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get2args(user_ctxt,
				      (unsigned int*)& user_str,
				      (unsigned int*)& len);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &path, user_str, len, 0);
	if (SOS_OK != retval)
	  break;

	retval = sos_fs_unlink(proc,
			       path, len);
	sos_kfree((sos_vaddr_t)path);
      }
      break;

    case SOS_SYSCALL_ID_SYMLINK:
      {
	sos_uaddr_t user_path, user_targetpath;
	sos_size_t  pathlen, targetpathlen;
	char * kernel_path;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get4args(user_ctxt,
				      (unsigned int*)& user_path,
				      (unsigned int*)& pathlen,
				      (unsigned int*)& user_targetpath,
				      (unsigned int*)& targetpathlen);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &kernel_path,
				      user_path,
				      pathlen, 0);
	if (SOS_OK != retval)
	  break;

	retval = sos_fs_symlink(proc,
				kernel_path, pathlen,
				user_targetpath, targetpathlen);
	sos_kfree((sos_vaddr_t)kernel_path);
      }
      break;

    case SOS_SYSCALL_ID_MKDIR:
      {
	sos_uaddr_t user_str;
	sos_size_t  len;
	sos_ui32_t  access_rights;
	char * path;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get3args(user_ctxt,
				      (unsigned int*)& user_str,
				      (unsigned int*)& len,
				      (unsigned int*)& access_rights);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &path, user_str, len, 0);
	if (SOS_OK != retval)
	  break;

	retval = sos_fs_mkdir(proc,
			      path, len, access_rights);
	sos_kfree((sos_vaddr_t)path);
      }
      break;

    case SOS_SYSCALL_ID_RMDIR:
      {
	sos_uaddr_t user_str;
	sos_size_t  len;
	char * path;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get2args(user_ctxt,
				      (unsigned int*)& user_str,
				      (unsigned int*)& len);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &path, user_str, len, 0);
	if (SOS_OK != retval)
	  break;

	retval = sos_fs_rmdir(proc, path, len);
	sos_kfree((sos_vaddr_t)path);
      }
      break;

    case SOS_SYSCALL_ID_CHMOD:
      {
	sos_uaddr_t user_str;
	sos_size_t  len;
	sos_ui32_t  access_rights;
	char * path;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get3args(user_ctxt,
				      (unsigned int*)& user_str,
				      (unsigned int*)& len,
				      (unsigned int*)& access_rights);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &path, user_str, len, 0);
	if (SOS_OK != retval)
	  break;

	retval = sos_fs_chmod(proc, path, len, access_rights);
	sos_kfree((sos_vaddr_t)path);
      }
      break;

    case SOS_SYSCALL_ID_STAT64:
      {
	sos_uaddr_t user_str;
	sos_size_t  len;
	sos_uaddr_t user_stat_struct;
	struct sos_fs_stat kernel_stat_struct;
	int nofollow;
	char * path;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get4args(user_ctxt,
				      (unsigned int*)& user_str,
				      (unsigned int*)& len,
				      (unsigned int*)& nofollow,
				      (unsigned int*)& user_stat_struct);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &path, user_str, len, 0);
	if (SOS_OK != retval)
	  break;

	retval = sos_fs_stat(proc, path, len, nofollow, & kernel_stat_struct);
	sos_kfree((sos_vaddr_t)path);
	if (SOS_OK != retval)
	  break;

	if (sizeof(kernel_stat_struct)
	    != sos_memcpy_to_user(user_stat_struct,
				  (sos_vaddr_t) & kernel_stat_struct,
				  sizeof(kernel_stat_struct)))
	  retval = -SOS_EFAULT;
      }
      break;

    case SOS_SYSCALL_ID_CHROOT:
    case SOS_SYSCALL_ID_CHDIR:
      {
	sos_uaddr_t user_str;
	sos_size_t  len;
	char * path;
	struct sos_fs_opened_file * of, * old_of;
	struct sos_process * proc;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get2args(user_ctxt,
				      (unsigned int*)& user_str,
				      (unsigned int*)& len);
	if (SOS_OK != retval)
	  break;

	retval = sos_memdup_from_user((sos_vaddr_t*) &path, user_str, len, 0);
	if (SOS_OK != retval)
	  break;

	retval = sos_fs_open(proc,
			     path, len,
			     SOS_FS_OPEN_DIRECTORY,
			     SOS_FS_OPEN_READ,
			     & of);
	sos_kfree((sos_vaddr_t)path);
	if (SOS_OK != retval)
	  break;

	if (syscall_id == SOS_SYSCALL_ID_CHROOT)
	  retval = sos_process_chroot(proc, of, & old_of);
	else
	  retval = sos_process_chdir(proc, of, & old_of);

	if (retval < 0)
	  {
	    sos_fs_close(of);
	    break;
	  }

	sos_fs_close(old_of);
      }
      break;      

    case SOS_SYSCALL_ID_FCHDIR:
      {
	struct sos_fs_opened_file * of, * new_of, * old_of;
	struct sos_process * proc;
	int fd;

	proc = sos_thread_get_current()->process;
	retval = sos_syscall_get1arg(user_ctxt,
				     (unsigned int*)& fd);
	if (SOS_OK != retval)
	  break;

	of = sos_process_get_opened_file(proc, fd);
	if (NULL == of)
	  {
	    retval = -SOS_EBADF;
	    break;
	  }

	/* Duplicate this FD */
	retval = sos_fs_duplicate_opened_file(of, proc, & new_of);
	if (SOS_OK != retval)
	  break;

	/* Do the actual chdir */
	retval = sos_process_chdir(proc, new_of, & old_of);
	if (retval < 0)
	  {
	    sos_fs_close(new_of);
	    break;
	  }

	sos_fs_close(old_of);
      }
      break;

    case SOS_SYSCALL_ID_BOCHS_WRITE:
      {
	sos_uaddr_t user_str;
	sos_size_t len;
	char * str;
	retval = sos_syscall_get2args(user_ctxt, & user_str, & len);
	if (SOS_OK != retval)
	  break;

	retval = sos_strndup_from_user(& str, user_str, len + 1, 0);
	if (SOS_OK == retval)
	  {
	    sos_bochs_printf("THR 0x%x: ",
			     (unsigned)sos_thread_get_current());
	    retval = sos_bochs_putstring(str);
	    retval = len;
	    sos_kfree((sos_vaddr_t)str);
	  }
      }
      break;


      /* ***********************************************
       * Debug syscalls (will be removed in the future)
       */


      /**
       * Syscall 4012: hexdump of a user-space memory region
       * args: addr_start size, retval=ignored
       */
    case 4012:
      {
	sos_uaddr_t user_str;
	unsigned int len;
	unsigned char * str;

	retval = sos_syscall_get2args(user_ctxt, & user_str, & len);
	if (SOS_OK != retval)
	  break;

	str = (char*)sos_kmalloc(len, 0);
	if (str)
	  {
	    int i;
	    sos_bochs_printf("Hexdump(0x%x, %d):\n", user_str, len);
	    retval = sos_memcpy_from_user((sos_vaddr_t) str, user_str, len);
	    sos_bochs_printf("  (Successfully copied %d out of %d)\n",
			     retval, len);

	    for (i = 0 ; i < retval ; i++)
	      {
		if ((i % 32) == 0)
		  sos_bochs_printf("%x:", i);
		sos_bochs_printf(" %x", str[i]);
		if (((i+1) % 32) == 0)
		  sos_bochs_printf("\n");
	      }
	    if (i % 32)
	      sos_bochs_printf("\n");

	    sos_kfree((sos_vaddr_t)str);
	  }
	else
	  retval = -SOS_ENOMEM;
      }
      break;

      
      /**
       * Syscall 4004: lists the VR of the current thread's address space
       * args: debug_string, retval=ignored
       */
    case 4004:
      {
	sos_uaddr_t ustr;
	char * kstr;
	struct sos_umem_vmm_as * my_as;

	retval = sos_syscall_get1arg(user_ctxt, & ustr);
	if (SOS_OK != retval)
	  break;

	retval = sos_strndup_from_user(& kstr, ustr, 256, 0);
	if (SOS_OK != retval)
	  break;

	extern void sos_dump_as(const struct sos_umem_vmm_as *, const char *);
	my_as
	  = sos_process_get_address_space(sos_thread_get_current()->process);
	sos_dump_as(my_as, kstr);
	sos_kfree((sos_vaddr_t)kstr);
      }
      break;


      /**
       * Syscall 4008: dump the list of processes in the system
       * args: none, retval=ignored
       */
    case 4008:
      {
	extern void sos_process_dumplist(void);
	sos_process_dumplist();
	retval = SOS_OK;
      }
      break;

    default:
      sos_bochs_printf("Syscall: UNKNOWN[%d]\n", syscall_id);
      retval = -SOS_ENOSUP;
    }
  
  return retval;
}
